/*
						 ***********
						** Summary **
						 ***********
	(1)	Title:
				TES:2 Daggerfall Saved Game Streams

	(2)	Purpose:
			To access the contents of a Daggerfall saved game.

	(3)	Non-ANSI features:
*/
#if defined(__GNUC__)
#	pragma pack(1)
#endif
/*
	(4)	Platform dependencies:
			The ostrstream pcount() function is not portable between GNU C++
			and Borland C++.

	(5)	Change history:
			Please see the RCS logs.

	(6)	Input:

			Classes:
				***  None  ***

			Common:
				***  None  ***

			Files:
				savetree.dat		--	The saved game archive to browse.
				savevars.dat		--	The faction relationships to the PC.

			Assertions about input:
				***  None  ***

	(7)	Output:

			Classes:
				***  None  ***

			Common:
				SaveTreeTypes		--	Names for the chunks.

			Files:
				savetree.dat		--	The updated saved game archive.

			Assertions about the output:
				***  None  ***

	(8)	Public members:

		1.	Title:
					ostream& ColumnStream::advance

			Purpose:
				To advance the output stream to the next column of output.  If
				we are already at the last column of the stream, advance to the
				next line.

			Parameters:
				return				--	Answer the column stream object.


		2.	Title:
					ostream& ColumnStream::flushColumn

			Purpose:
				To flush the formatted text in the current output column onto
				the associated output stream for display.

			Parameters:
				return				--	Answer the column stream object.


		3.	Title:
					Arena2ObjectData* SaveTree::findItem

			Purpose:
				To answer the next data object from the saved game file.

			Parameters:
				return				--	Answer the located item record, or nil
										if no item could be located.


		4.	Title:
					int SaveTree::forEach

			Purpose:
				To iterate over each chunk record in the saved game, applying
				the specified function to each record.

			Parameters:
				ChunkIterator* item	--	The function to apply to each record.
				return				--	Answer true if the operation was
										successful, and false otherwise.


	(9)	Private members:

		1.	Title:
					ostream& dumpBytes

			Purpose:
				To render a string of bytes in hexadecimal on an output stream.

			Parameters:
				ostream& os			--	The output stream to write on.
				char* cPtr			--	The byte stream to translate.
				int count			--	The number of bytes to convert.
				return				--	Answer the output stream.


		2.	Title:
					ostream& dumpShorts

			Purpose:
				To render a string of bytes in hexadecimal on an output stream.
				The bytes are expressed as 16-bit integers.

			Parameters:
				ostream& os			--	The output stream to write on.
				char* cPtr			--	The byte stream to translate.
				int count			--	The number of half-words to convert.
				return				--	Answer the output stream.


		3.	Title:
				Chunk* SaveTree::chunk

			Purpose:
				Answer the next chunk from the save file.  Answer nil if
				the end of file occurs.  Leave the file positioned after
				the current chunk, i.e., at the start of the next chunk
				header.

			Parameters:
				Chunk* chunk		--	The incomplete chunk information.  On
										entry, the chunkSize member specifies
										the number of consecutive bytes to read
										from the save game file.  If the chunk
										member is not nil, it will be deleted.
				return				--	Answer the completed chunk data, or nil
										on error.


		4.	Title:
					ChunkHeader* SaveTree::nextHeader

			Purpose:
				Obtain the next chunk header.  If the next chunk has 0 bytes,
				advance to the next chunk with a non-zero size.  Answer nil
				if the end of file occurs.  Leave the file positioned at the
				beginning of the current chunk.  Use the existing header size
				to compute the offset to the next header.

			Parameters:
				return				--	Answers the next header field from
										from the save game file or nil at
										end-of-file.


		5.	Title:
					int SaveTree::findChunk

			Purpose:
				Position the input stream at the beginning of the next item
				record.  Answer false if no item was found, and true otherwise.

			Parameters:
				return				--	Answer true if an item was found, and
										0 otherwise.


	(10) Subprograms called:
			***  None  ***

	(11) Includes referenced:
			chunk.h					--	The Daggerfall save game prototypes.

	(12) Classes declared:
			***  None  ***

						 ****************
						** Declarations **
						 ****************

					 ***********************
					** Global declarations **
					 ***********************

				  Copyright 1999, Nagual Enterprises
			Please see COPYING for the complete GNU license

					  ********************
					 ** Global functions **
					  ********************
*/

#			ifndef __BORLANDC__
#				include <unistd.h>
#			else
#				include <stdlib.h>
#			endif
#				include <errno.h>
#				include <string.h>
#				include <sys/stat.h>
#				include "chunk.h"

/*
					  *************
					 ** Constants **
					  *************
*/

/*
					   **********
					  ** Macros **
					   **********
*/

/*
					  ***********
					 ** Classes **
					  ***********
*/

/*
					   **********
					  ** Common **
					   **********
*/
					int SaveTree::OpenMode = 0;
					int SaveTree::NullCounts = 0;

					/*	Names of the savetree.dat chunk types.  NULL
						implies never before seen. */

				char* SaveTreeTypes[256]={
					"Unknown_00", NULL, "Item", "Player",
					"Player parent 1", "Player parent 2", "Unknown_06", "Underground?",
					"Unknown_08", "Spell", "Guild", "Stat boni",
					NULL, NULL, "Qbn image", NULL,
					"Qbn parent 1", NULL, "Qbn foe", NULL,
					NULL, NULL, "Magic foes", "Control panel",
					"Logbook", "Bank accounts", NULL, NULL,
					"Original class", "Unknown_1d", NULL, "Potion",
					"Location", "Temp. treasure pile", "Creature", NULL,
					"Store item", NULL, NULL, "Unknown_27",
					"Qbn place", "Qbn person", "Unknown_2a", "Unknown_2b",
					"Creature", "NPC", "Generic NPC", NULL,
					NULL, NULL, NULL, "Current *.amf",
					"Container", "Unknown_35", NULL, NULL,
					NULL, NULL, "Treasure pile", NULL,
					NULL, NULL, NULL, NULL,
					"Visited?", "Qbn questor", NULL, NULL,
					NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
					NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
				};


/*
				 **********************
				** Local declarations **
				 **********************
*/

					/*	Placate forward references for the compiler. */

				ostream& dumpBytes(	ostream& os,
									char* cPtr,
									int count);
				void pError(const char* txt);

/*
					  *************
					 ** Variables **
					  *************
*/

						/*	Token where player changes commence. */

					char** cliIndex = NULL;
//					fstream qbn("player.qbn", forBinaryIO | ios::trunc);
					fstream qbn;
					char text[1024];
					ColumnStream line(text, 1024);

/*
						 ********
						** Text **
						 ********


							 *******************
							** Private methods **
							 *******************
*/

ostream& dumpBytes
(	ostream& os,
	char* cPtr,
	int count)
{
		/*	Dump the first count bytes of the current chunk on a stream. */

	int code;

	os << hex;
	while (count--)
	{
		code = (int)*cPtr++;
		code &= 0xff;
		os
			<< setw(2) << setfill('0')
			<< code;
	}
	return os << dec; // << endl;
}

//void dumpShorts
//(	ostream& os,
//	char* cPtr,
//	int count)
//{
		/*	Dump the first count half-words of the current chunk on a stream. */

//	int code, codeLo;

//	os << hex;
//	while (count--)
//	{
//		codeLo = (int)*cPtr++;
//		codeLo &= 0xff;
//		code = (int)*cPtr++;
//		code &= 0xff;
//		os
//			<< setw(2) << setfill('0')
//			<< code
//			<< setw(2) << setfill('0')
//			<< codeLo
//			<< ' ';
//	}
//	os << dec;
//}

Chunk* SaveTree::chunk
(	Chunk* chunk)
{
		/*	Answer the next chunk from the save file.  Answer nil if the end
			of file occurs.  Leave the file positioned after the current chunk,
			i.e., at the start of the next chunk header. */

	if (chunk->chunk)
	{
		delete [] chunk->chunk;
		chunk->chunk = NULL;
	}
	if (_handle->eof())
		return NULL;

	chunk->chunk = new char[chunk->chunkSize];
	if (!chunk->chunk)
	{
		pError("chunk");
		cout
			<< "chunk: new char["
			<< chunk->chunkSize
			<< "] has failed"
			<< punt;
		return NULL;
	}
	_handle->read((char*)chunk->chunk, chunk->chunkSize);
	if (_handle->gcount() == (int)chunk->chunkSize)
		return chunk;
	return NULL;
}

Chunk* SaveTree::nextChunk
(	void)
{
		/*	Answer the next chunk record, or nil at end of file. */

	if (!_isOpen)
		/*	Open the saved game file since this is the first request. */

		openChunk();

	/*	Position the file to the next item chunk. */

	if (_handle->eof())
		return NULL;
	if (!nextHeader())
		return NULL;
	_offset = _handle->tellg();
	_currentChunk.chunkSize = _header.chunkSize;
	if (!chunk(&_currentChunk))
		/*	We've found the end-of-file instead. */

		return NULL;

	/*	Remember the fact we've already positioned the file stream by
		forgetting the record size. */

	_header.chunkSize = 0;
	return &_currentChunk;
}

//Chunk* SaveTree::findChunk
//(	unsigned char type)
//{
		/*	Answer the next chunk record matching the specified chunk type,
			or nil at end of file. */

//	if (!_isOpen)
		/*	Open the saved game file since this is the first request. */

//		openChunk();

//	while (_header.type != type)
//	{
		/*	Position the file to the next item chunk. */

//		if (_handle->eof())
//			return NULL;
//		if (!nextHeader())
//			return NULL;
//		_offset = _handle->tellg();
//		if (_header.type == type)
//		{
//			_currentChunk.chunkSize = _header.chunkSize;
//			if (!chunk(&_currentChunk))
//				/*	We've found the end-of-file instead. */

//				return NULL;

			/*	Remember the fact we've already positioned the file stream by
				forgetting the record size. */

//			_header.chunkSize = 0;
//			return &_currentChunk;
//		}
//	}
//	return NULL;
//}

ChunkHeader* SaveTree::nextHeader
(	void)
{
		/*	Obtain the next chunk header.  If the next chunk has 0 bytes,
			advance to the next chunk with a non-zero size.  Answer nil
			if the end of file occurs.  Leave the file positioned at the
			beginning of the current chunk.  Use the existing header size
			to compute the offset to the next header. */

	if (_handle->eof())
		/*	We done. */

		return NULL;
	if (_header.chunkSize)
		/*	Skip over the current chunk. */

		_handle->seekg(_header.chunkSize, ios::cur);
	while (1)
	{
		_header.chunkSize = 0;
		_handle->read((char*)&_header, sizeof(ChunkHeader));
		if (_handle->eof())
			break;
		if (_handle->gcount() != sizeof(ChunkHeader))
			break;
		_handle->seekg(-1L, ios::cur);

		if (_header.chunkSize && NullCounts > 0)
			return &_header;
		if (!NullCounts++)
		{
			/*	Always ignore the first chunk, whether populated
				or not. */

			if (_header.chunkSize)
				_handle->seekg(_header.chunkSize, ios::cur);
			continue;
		}
		if (NullCounts >= 3)
			/*	Stop reading after the 3rd null chunk regardless. */

			break;

		/*	Empty chunks punctuate the save game into divisions or
			sections.  Every well-formed savetree seems to have
			3 empty chunk punctuations.  The 3rd empty seems to
			indicate the end of the chunk-formatted entries in
			savetree.dat. */

//		if (ReportNulls)
//			cout
//				<< "-- 0x"
//				<< setbase(16)
//				<< _handle->tellg() - 4
//				<< ": found empty chunk " << setw(2) << setfill('0')
//				<< setbase(0)
//				<< NullCounts
//				<< endl;
	}
	return NULL;
}

int SaveTree::openChunk
()
{
		/*	Open the saved game file if this is the first request. */

	if (!_isOpen)
	{
		_handle->open(_name, OpenMode);
		if (isFileErrorp(_handle))
		{
			pError(_name);
			_exit(3);
		}
		_handle->seekg(0x13, ios::beg);
		_currentChunk.chunkSize = _header.chunkSize = 0;
		_currentChunk.chunk = NULL;
		_header.type = 0;

		/*	Remember we've already done our job for the next request. */

		_isOpen = 1;
	}
	return 1;
}


/*
							 ******************
							** Public methods **
							 ******************
*/

void pError
(	const char* txt)
{
		/*	Gripe upon the current anomaly in the system. */

	cerr
		<< txt
		<< '('
		<< errno
		<< ") "
		<< (errno ?
			strerror(errno) :
			"Griping for no apparent reason.\n");
}

/*
					 *****************
					** Class methods **
					 *****************
*/

/*
					 ********************
					** Instance methods **
					 ********************
*/


ostream& ColumnStream::advance
(	void)
{
		/*	Advance the output stream to the beginning of the next column
			by padding the current column with blanks up to the next column
			boundary. */

	int charCount;
	int padChars;
#if defined(__BORLANDC__)
	char blanks[256];
#else
	char blanks[columnWidth+1];
#endif

	/*	Get the string length, ignoring the terminal newline, and figure out
		how many padding blanks are required.  Note that the intrinsic
		pcount() will return different values according to the platform
		implementing it.  So we use tellp() instead. */

	charCount = basis.tellp() - 1;
	padChars = columnWidth - ((charCount % columnWidth) + 1);

	/*	Toast the newline and deduct that from the padding cost.  Scroll the
		stream pointer back one position for good measure. */

	associatedBuf[charCount] = ' ';
	padChars--;
	basis.seekp(-1, ios::cur);

	/*	Build the padding string. */

	if (padChars > 0)
	{
		memset(blanks, ' ', padChars);
		blanks[padChars] = '\0';
		basis /* *this*/ << blanks;
	}
	return currentColumn++ >= columnCount ?
		flushColumn() :
		(ostream&)basis /**this*/;
}

ostream& ColumnStream::flushColumn
(	void)
{
		/*	Flush the column onto the associated output stream. */

	basis /* *this*/ << '\0';
	basis.flush();
	associatedStream
		<< associatedBuf
		<< endl;
	rewind();
	return (ostream&)basis /* *this*/;
}

int SaveTree::forEach
(	ChunkIterator* item)
{
		/*	For each chunk record in the saved game, operate on
			its data with each operator in the chunk iterator list. */

	int reply = true;
	Chunk* object;

	while (1)
	{
		/*	Fetch a chunk record. */

		object = nextChunk();
		if (!object)
			/*	Found the end-of-file instead.  Answer normal completetion. */

			return 1;

		ChunkIterator* options=item;

			/*	Borland chokes on the obvious expression (*options), but accepts
				the direct index.  Go figure. */

		while (options[0] != NULL)
		{
			if (!(reply = (_browser->*(*options++))(object)))
				return 0;
		}
		if (!reply)
			break;
	}
	return 0;
}

int SaveTree::update
(	long offset,
	Chunk* aChunk)
{
		/*	Update (in place) the specified chunk segment.  Answer true
			if the update has succeeded and false otherwise. */

	_handle->clear();
	_handle->seekp(offset, ios::beg);
	_handle->write((char*)aChunk->chunk, aChunk->chunkSize);
	return !isFileErrorp(_handle);
}

