/*
								 ***********
								** Summary **
								 ***********

					  Copyright 1998, Nagual Enterprises
				Please see COPYING for the complete GNU license

	(1)	Title:
				File shorthands for ios and Daggerfall quest files.

	(2)	Purpose:
			To provide helpful mnemonics for various aspects of file
			access in Daggerfall.

	(3)	Change history:
			Please see the RCS logs.

	(4)	Definitions:

			Constants:
				_asm				--	ANSI-fication for Borland C/C++.
				_file				--	Include already processed flag.
				DirSeperator		--	Pathname component delimiter.
				false				--	The universal false value.
				forBinaryIO			--	ios flags for binary input/output.
				forBinaryI			--	ios flags for binary read-only input.
				forTextI			--	ios flags for text read-only input.
				forTextIO			--	ios flags for text input/output.
				forTextO			--	ios flags for text output only.
				logInt				--	Log (base 2) of the number of bytes / int.
				logLong				--	Log (base 2) of the number of bytes / long.
				logShort			--	Log (base 2) of the number of bytes / short.
				true				--	The universal true value.

			Macros:
				isEndOfFile(z)		--	Answer true if (z) has reached file-end.
				isEndOfFilep(z)		--	Answer true if (*z) has reached file-end.
				isEquals(x,y)		--	Answer true if two strings match.
				isEqualIs(x,y)		--	Answer true if two strings match
										regardless of case.
				isFileError(z)		--	Answer true if (z) has a file error.
				isFileErrorp(z)		--	Answer true if (*z) has a file error.
				isNil(z)			--	Answer true if (z) is nil.
				isSlotValue(z)		--	Answer true if (z) is a proper index.
				notFinalSegment(z)	--	Answer true if (z) is not the last
										piece of the save game file.
				notNil(z)			--	Answer true if (z) is not nil.

			Classes:
				QuestFileGuild		--	Index of quest-giver organizations.
				QuestFileStatus		--	Index of questor relationship to the
										quest-giving guild.
				QuestFileSubtype	--	Index of questor ranking to receive
										a quest.
				QuestFileMethod		--	Index of ways to start a quest.

			Assertions about definitions:
				Note that BorlandC gags with *nix style line terminators
				and turns off Ansi preprocessing for some reason.
				Replacing #if defined() with the older style #ifdef
				placates its brokeness.  (Actually, the trigger for
				the misbehavior is combining *nix style headers with
				msdos style.  It forgets to switch back and forth
				between the alternate <CR> present/absent and so
				stuffs its LALR rather badly)

	(5)	References:

			Include files:
				***  None  ***

			Methods:
				***  None  ***

			Classes:
				***  None  ***

			Assertions about the references:
				***  None  ***

								 ********
								** Text **
								 ********
*/

#ifndef _file
#	define _file

#	include <iostream.h>
#	include <fstream.h>
#	include <strstream.h>
#	include <iomanip.h>

#	define isNil(z)				(!(z))
#	define notNil(x)			(!isNil(x))
#	define isFileError(z)		((z).bad())
#	define isFileErrorp(x)		((x)->bad())
#	define isEquals(x, y)		(!::strcmp(x, y))
#	define isEqual(x, y, z)		(!::strncmp(x, y, z))
#	ifdef __BORLANDC__
#		define DirSeperator			'\\'
#		define isEqualIs(x, y)		(!::stricmp(x, y))
#		define isEqualI(x, y, z)	(!::strncmpi(x, y, z))
#		define BitBucket			"nul"
#	else
#		define DirSeperator			'/'
#		define isEqualIs(x, y)		(!::strcasecmp(x, y))
#		define isEqualI(x, y, z)	(!::strncasecmp(x, y, z))
#		define BitBucket			"/dev/null"
#	endif
#	define true					1
#	define false				0
#	define forBinaryIO			(ios::in | ios::out | ios::binary)
#	define forBinaryI			(ios::in | ios::binary)
#	define forTextIO			(ios::in | ios::out | ios::ate)
//#	define forTextI				(ios::in | ios::ate)
#	define forTextI				(ios::in)
#	define forTextO				(ios::out | ios::trunc)
#	define notFinalSegment(z)	((z)->type != (unsigned int)EOF)
//#	define isSlotValue(x)		(isNil(::strchr((x), '.')))
#	define isSlotValue(z)		isdigit(*(z))
#	define _asm					asm
#	define isEndOfFile(z)		((z).eof())
#	define isEndOfFilep(z)		((z)->eof())

		/*	Define some convenience constants for index mapping. */

#	ifdef __BORLANDC__
#		define logInt	1
#	else
#		define logInt	2
#	endif
#	define logShort		1
#	define logLong		2

	//	Quest file name format

	//	Byte 0:

	enum QuestFileGuild {
		tutorial=95,		//	'_'
		cureDisease=36,		//	'$'
		julianos=48,		//	'0'
		meridia,			//	'1'	Daedra prince quests
		molagBal,			//	'2'
		namira,				//	'3'
		nocturnal,			//	'4'
		peryite,			//	'5'
		sheogorath,			//	'6'
		sanguine,			//	'7'
		malacath,			//	'8'
		peasant=65,			//	'A' Guildly quests
		knightly,			//	'B'
		temple,				//	'C'
		akatosh,			//	'D'
		arkay,				//	'E'
		dibella,			//	'F'
		kynaran,			//	'G'
		mara,				//	'H'
		stendarr,			//	'I'
		zenithar,			//	'J'
		merchant,			//	'K'
		brotherhood,		//	'L'
		fighter,			//	'M'
		mage,				//	'N'
		thief,				//	'O'
		vampire,			//	'P'
		coven,				//	'Q'
		royalty,			//	'R'
		mainStory,			//	'S' The Main Story
		ladyAzura,			//	'T' More daedra prince quests
		boethiah,			//	'U'
		clavicusVile,		//	'V'
		hermaeusMora,		//	'W'
		hircine,			//	'X'
		mehruneDagon,		//	'Y'
		mephala				//	'Z'
	};

	//	Byte 1: always zero (0).

	//	Byte 2:

	enum QuestFileStatus {	joiningGuildQuest=65,	//	'A'
							memberOnlyQuest,		//	'B'
							anyonesQuest};			//	'C'

	//	Byte 3,4: [0,8] for the minimum guild rank required to receive this
	//			quest.  Also indicates the required reputation with the
	//			quest grantor.  'X' apparently means '?' or any digit.
	//			'a' is a mystery in o0a0aL00, which is apparently the
	//			invitation to join quest.

	//	Byte 4:

	enum QuestFileSubtype {	stock=48,				//	'0'
							alt1,					//	'1'
							alt2=65,				//	'A'
							alt3=88};				//	'X'

	//	Byte 5:

	enum QuestFileMethod {	byLetter=76,			//	'L'
							inPerson=89};			//	'Y'

	//	Byte 6,7:  unique quest number.

#endif					/*	_file */
