/*
						 ***********
						** Summary **
						 ***********
	(1)	Title:
				Thaw Daggerfall Main Quest

	(2)	Purpose:
			To activate stalled subplots of the main quest in TES:2.

	(3)	Non-ANSI features:
*/
#if defined(__GNUC__)
#	pragma pack(1)
#endif
/*
	(4)	Platform dependencies:
			Note the use of _exit() to work around brokeness in glibc.

	(5)	Change history:
			***  None  ***

	(6)	Input:

			Classes:
				SaveGameBrowser			--	Generic savetree brower.

			Parameters:
				***  None  ***

			Common:
				***  None  ***

			Files:
				savetree.dat			--	Saved game to modify.

			Assertions about input:

				Syntax:
					thaw  slot#  [Aubk-i	|
								  Cyndassa	|
								  Elysanna	|
								  Helseth	|
								  Lhotun	|
								  Medora	|
								  Morgiah]

					where slot# is the saved game slot to edit in [0,5]
					and the following parameter is the name of the subplot
					to activate.

	(7)	Output:

			Classes:
				***  None  ***

			Parameters:
				***  None  ***

			Common:
				***  None  ***

			Files:
				savetree.dat			--	Modified savetree.dat.

			Assertions about the output:
				The saved game file is updated in place.  No
				automatic backup is created.

	(8)	Public members:

		1.	Title:
					int main

			Purpose:
				Enable/disable the specified quest event.

			Parameters:
				int argc			--	Count of cli tokens.
				char** argv			--	List of cli tokens.
				return				--	Answer 0 on success, and not 0
										otherwise.


		2.	Title:
					int QuestTrigger::changeEvent

			Purpose:
				Enable the respective main quest task for the PC's current
				affliction.

			Parameters:
				return				--	Answer true on success, and
										false otherwise.


		3.	Title:
					int QuestTrigger::snagOrgClass

			Purpose:
				Catch the original character class, if any.

			Parameters:
				Chunk* chunk		--	Current chunk to examine for
										a match.
				return				--	Answer true to continue processing
										and false otherwise.


		4.	Title:
					int QuestTrigger::snagOrgStats

			Purpose:
				Catch the original character stats, if any.

			Parameters:
				Chunk* chunk		--	Current chunk to examine for
										a match.
				return				--	Answer true to continue processing
										and false otherwise.


		5.	Title:
					int QuestTrigger::snagQbn

			Purpose:
				Catch the chunk record matching this edit request.

			Parameters:
				Chunk* chunk		--	Current chunk to examine for
										a match.
				return				--	Answer true to continue processing
										and false otherwise.


		6.	Title:
					QuestTrigger::~QuestTrigger

			Purpose:
				To release any snagged Chunk memory when an instance
				of the class is deleted.

			Parameters:
				***  None  ***


	(9)	Private members:

		1.	Title:
					inline void QuestTrigger::abscond

			Purpose:
				To transfer ownership of a chunk specimen.

			Parameters:
				Chunk* mine			--	The chunk claiming ownership.
				Chunk* org			--	The chunk relinquishing ownership.


	(10) Subprograms called:
			***  None  ***

	(11) Includes referenced:
			browser.h				--	Generic savetree browser.
			chunk.h					--	Savetree disk access methods.

	(12) Classes declared:
			QuestTrigger			--	Quest editor for s0000999.

						 ****************
						** Declarations **
						 ****************

					 ***********************
					** Global declarations **
					 ***********************

				  Copyright 2001, Nagual Enterprises
			Please see COPYING for the complete GNU license

					  ********************
					 ** Global functions **
					  ********************
*/

#			ifndef __BORLANDC__
#				include <unistd.h>
#			endif
#				include <stdlib.h>
#				include <errno.h>
#				include <string.h>
#				include <sys/stat.h>
#				include <ctype.h>
#				include "chunk.h"
#				include "browser.h"

				extern void pError(const char*);

/*
					  *************
					 ** Constants **
					  *************
*/

/*
					   **********
					  ** Macros **
					   **********
*/

					/*	Kludge to avoid the glibc strstream
						brokeness if return() or exit() are
						used from main().  This means all stream
						i/o must be flushed manually before
						closing the stream, else bits will fall
						onto the floor. */

#				ifdef __BORLANDC__
#					define Die(a)	return a
#				else
#					define Die(a)	_exit(a)
#				endif

/*
					  ***********
					 ** Classes **
					  ***********
*/

					/*	Define a class to (re)start subplots
						in the main quest. */

					class QuestTrigger : public SaveGameBrowser
					{
					protected:
						int _foundActive;		//	Detected subplot active.
						int _found998;			//	Detected wolf scheduler.
						int _found997;			//	Detected boar scheduler.
						int _found996;			//	Detected vamp scheduler.
						long _qbnOffset;		//	Qbn image in savetree.dat
						Chunk _qbn;				//	The qbn image to edit.
						Chunk _originalClass;	//	Original class record.
						Chunk _originalStats;	//	Original PC stats.

						void abscond(Chunk* mine,
									 Chunk* c);

					public:
						QuestTrigger(SaveTree* s,
									 char** ci) :
							SaveGameBrowser(s, ci),
							_foundActive(0),
							_found998(0),
							_found997(0),
							_found996(0),
							_qbnOffset(-1) {s->browser(this);};
						~QuestTrigger();

						int changeEvent();		//	Edit stalled quest.
						int snagQbn(Chunk* c);	//	Snag s0000999 image.
												//	Snag original class.
						int snagOrgClass(Chunk* c);
												//	Snag original stats.
						int snagOrgStats(Chunk* c);
						static const char copyRight[];
					};

					const char QuestTrigger::copyRight[]=
						"thaw: Daggerfall Quest Revival, "
						"version 1.0"
						"\n      Copyright (c) 2001, Nagual Enterprises";

/*
					   **********
					  ** Common **
					   **********

					***  None  ***

					 **************
					** Parameters **
					 **************


				 **********************
				** Local declarations **
				 **********************

					  *************
					 ** Variables **
					  *************
*/

					/*	Chunk iterators used to thaw a quest. */

					typedef int (QuestTrigger::*QuestTriggerIterator)(Chunk*);
					static QuestTriggerIterator chunkActions[]={
						&QuestTrigger::snagQbn,
						&QuestTrigger::snagOrgClass,
						&QuestTrigger::snagOrgStats,
						NULL,
						NULL
					};

					/*	Thawable subplot names. */

					static char* plotters[]={
						"Aubk-i",
						"Cyndassa",
						"Elysanna",
						"Helseth",
						"Lhotun",
						"Medora",
						"Morgiah",
						NULL
					};
					/*	Sets of task edit lists.  Each pair of
						entries is a task number, new task flag
						for each of the tasks involved in a particular
						subplot.  The pair-list is terminated by a
						-1 task number.  The fourth pair of entries
						must be the task number of the event block
						that schedules the subplot in question. */

					/*	Task edits for the orignal BS quest task
						number scheme. */

					static short oldAubkiList[]={
						11, 0, 16, 1, 4, 1, 17, 0, -1
					};
					static short oldCyndiList[]={
						6, 0, 9, 1, 16, 1, 20, 0, -1,
					};
					static short oldEllyList[]={
						5, 0, 10, 1, 16, 1, 31, 0, 21, 0 -1
					};
					static short oldHelsethList[]={
						1, 0, 14, 1, 16, 1, 27, 0, 22, 0, -1
					};
					static short oldLhotunList[]={
						0, 0, 15, 1, 16, 1, 23, 0, -1
					};
					static short oldMedoraList[]={
						2, 0, 12, 1, 16, 1, 24, 0, -1
					};
					static short oldMorgiahList[]={
						3, 0, 12, 1, 16, 1, 25, 0, -1
					};
					static short* oldEvents[]={
						oldAubkiList,
						oldCyndiList,
						oldEllyList,
						oldHelsethList,
						oldLhotunList,
						oldMedoraList,
						oldMorgiahList,
					};
					/*	Task edits for the new (Nagual) task
						number scheme. */

					static short newAubkiList[]={
						10, 0, 19, 1, 4, 1, 22, 0, -1
					};
					static short newCyndiList[]={
						6, 0, 8, 1, 21, 1, 27, 0, -1
					};
					static short newEllyList[]={
						5, 0, 9, 1, 20, 1, 38, 0, 28, 0, -1
					};
					static short newHelsethList[]={
						1, 0, 13, 1, 16, 1, 34, 0, 29, 0, -1
					};
					static short newLhotunList[]={
						0, 0, 14, 1, 15, 1, 30, 0, -1
					};
					static short newMedoraList[]={
						2, 0, 12, 1, 17, 1, 31, 0, -1
					};
					static short newMorgiahList[]={
						3, 0, 11, 1, 18, 1, 32, 0, -1
					};
					static short* newEvents[]={
						newAubkiList,
						newCyndiList,
						newEllyList,
						newHelsethList,
						newLhotunList,
						newMedoraList,
						newMorgiahList,
					};

					static char activeQuests[] = "5762134";
					static char questName[] = "s0000000";
/*

						 ********
						** Text **
						 ********


							 *******************
							** Private methods **
							 *******************
*/

inline void QuestTrigger::abscond
(	Chunk* mine,
	Chunk* org)
{
		/*	Claim ownership of a chunk specimen. */

	mine->chunkSize = org->chunkSize;
	mine->chunk = org->chunk;
	org->chunk = NULL;
	return;
}

/*
							 ******************
							** Public methods **
							 ******************
*/

int QuestTrigger::changeEvent
()
{
		/*	Enable the requested main quest. */

	_qbn.chunkSize = 0;
	_qbn.chunk = NULL;
	abscond(&_originalClass, &_qbn);
	abscond(&_originalStats, &_qbn);
	int factionType = _cliIndex - plotters;
	questName[7] = activeQuests[factionType];

	/*	Browse the saved game chunks for the ones we're interested in. */

	_game->forEach((ChunkIterator*)chunkActions);

	/*	Sort out which version of the main quest is running in the
		saved game. */

	if (_foundActive)
	{
		cout
			<< "thaw: requested subplot ("
			<< *_cliIndex
			<< ") is already active.\n"
			<< "thaw: request failed."
			<< endl;
		return false;
	}
	if (_qbnOffset == -1)
	{
		cout
			<< "thaw: did not locate quest s0000999."
			<< endl;
		return false;
	}
	/*	Now, confirm we've got the right version(s) of the qbn file(s). */

	char* sectionStart = &((char*)_qbn.chunk)[sizeof(Arena2DataHeader)+6+10];
	short taskCount = ((short*)sectionStart)[9];
	short** eventList = taskCount < 47 ?
		oldEvents :
		newEvents;
	if (_found998 || _found997 || _found996)
	{
		/*	Sanity check. */

		if (eventList != newEvents)
		{
			cout
				<< "thaw: internal consistency check fails.\n"
				<< "      found quests s0000996 / s0000998 "
				<< "but quest s0000999 doesn't tie to those versions.\n"
				<< "thaw: request failed."
				<< endl;
			return false;
		}
	}
	/*	So far, so good.  Now flip the events for the desired subplot. */

	short offset = ((short*)sectionStart)[19];
	char* qbnStart  = &((char*)_qbn.chunk)[sizeof(Arena2DataHeader)];
	long* eventStart = (long*)&qbnStart[offset];
	short* events = eventList[factionType];

	/*	Walk the list of task edits, updating section 9 entries as
		decreed by the edit. */

	while (*events >= 0)
	{
		int base = (*events++) << 1;
		long value = ((long)*events++) << 24;
		eventStart[base] &= 0xffffff;
		eventStart[base] |= value;
	}
	/*	Now, flip the `executed' toggle for the subplot activation task. */

	events = eventList[factionType];
	int state = events[6];
	taskCount = ((short*)sectionStart)[8];
	offset = ((short*)sectionStart)[18];
	char* pCodes = (char*)&qbnStart[offset];

	/*	Walk the p-code records, looking for records matching the task
		number mandated by the edit and clearing their I've-been-executed
		flag. */

	while (taskCount--)
	{
		if (pCodes[13] == state)
			/*	Erase the `executed' marker for this instruction. */

			pCodes[2] = pCodes[3] = '\0';
		pCodes = &pCodes[87];
	}
	/*	Finally, update the editted qbn image in the saved game file. */

	if (_game->update(_qbnOffset, &_qbn))
	{
		cout
			<< "-- "
			<< *_cliIndex
			<< " subplot enabled."
			<< endl;
		return true;
	}
	pError("thaw");
	cout
		<< "thaw: could not enable "
		<< *_cliIndex
		<< " subplot."
		<< endl;
	return false;
}

int QuestTrigger::snagOrgClass
(	Chunk* chunk)
{
		/*	Catch the original character class, if any. */

	if (chunk->type() != ChunkOriginalStats)
		return true;
	abscond(&_originalStats, chunk);
	return true;
}

int QuestTrigger::snagOrgStats
(	Chunk* chunk)
{
		/*	Catch the original character stats, if any. */

	if (chunk->type() != ChunkOriginalClass)
		return true;
	abscond(&_originalClass, chunk);
	return true;
}

int QuestTrigger::snagQbn
(	Chunk* chunk)
{
		/*	Catch the chunk record matching this edit request. */

	if (chunk->type() != ChunkQuest)
		return true;
	char* qName = &((char*)chunk->chunk)[sizeof(Arena2DataHeader)+6];
	if (isEqualI(qName, questName, 8))
	{
		/*	Found a copy of the requested subplot already active. */

		_foundActive = true;
		return false;
	}
	if (isEqualI(qName, "s0000998", 8))
	{
		_found998 = true;
		return true;
	}
	if (isEqualI(qName, "s0000997", 8))
	{
		_found997 = true;
		return true;
	}
	if (isEqualI(qName, "s0000996", 8))
	{
		_found996 = true;
		return true;
	}
	if (!isEqualI(qName, "s0000999", 8))
		return true;
	abscond(&_qbn, chunk);
	_qbnOffset = _game->offset();
	return true;
}

QuestTrigger::~QuestTrigger
()
{
	if (_originalClass.chunk)
	{
		delete _originalClass.chunk;
		_originalClass.chunk = NULL;
	}
	if (_originalStats.chunk)
	{
		delete _originalStats.chunk;
		_originalStats.chunk = NULL;
	}
}

int usage
()
{
	cout
		<< "Usage:  thaw  slot#  aSubplotName\n\n"
		<< "where slot# is the number of the save game slot to use "
		<< "between 0 and 5.\n"
		<< "aSubplotName is the subplot to revive:\n"
		<< "    Aubk-i\n"
		<< "    Cyndassa\n"
		<< "    Elysanna\n"
		<< "    Helseth\n"
		<< "    Lhotun\n"
		<< "    Medora\n"
		<< "    Morgiah\n"
		<< endl
		<< QuestTrigger::copyRight
		<< endl;
	return 1;
}

int main
(	int argc,
	char** argv)
{
		/*	Enable/disable the specified quest event. */

	if (argc != 3)
	{
		if (argc != 1)
			cout
				<< "Wrong number of command line arguments."
				<< endl;
		Die(usage());
	}
	/*	Sanitize the save game slot number. */

	if (!isdigit(*argv[1]))
	{
		cout
			<< '"'
			<< argv[1]
			<< '"'
			<< " is not a valid save game slot number."
			<< endl;
		Die(usage());
	}
	int slot = strtol(argv[1], NULL, 0);
	if (slot < 0 || slot > 5)
	{
		cout
			<< '"'
			<< argv[1]
			<< '"'
			<< " save game slot not between 0 and 5."
			<< endl;
		Die(usage());
	}
	/*	Sanitize the subplot name. */

	char** plots = plotters;
	while (*plots)
	{
		if (isEqualIs(argv[2], *plots))
			break;
		plots++;
	}
	if (!*plots)
	{
		cout
			<< '"'
			<< argv[2]
			<< '"'
			<< " is an unknown subplot name."
			<< endl;
		Die(usage());
	}
	/*	Build the path name to the save game. */

	char infile[256];
	ostrstream itag(infile, 256);
	char dirPath[256];
	char* dPath = getenv("TES2");
	ostrstream dtag(dirPath, 256);
	if (dPath)
		dtag
			<< getenv("TES2")
			<< DirSeperator;
	dtag
		<< SavePrefix
		<< slot
		<< DirSep
		<< '\0';
	itag
		<< dirPath
		<< SaveFile
		<< '\0';

	/*	Let him know what we're trying to use. */

	cout
		<< "-- Using slot "
		<< slot
		<< " `"
		<< infile
		<< "' for input";
	cout << punt;

	/*	Verify the save file is present. */

	struct stat instat;
	if (stat(infile, &instat))
	{
		/*	Can't find the expected save game file.  Probably not running out
			of the game directory. */

		pError("thaw");
		cout
			<< "thaw: cannot access "
			<< infile
			<< punt;
		Die(usage());
	}
	/*	Tell him how big the save file is. */

	unsigned long fSize = fileSize(instat);
	cout
		<< "-- Savetree is "
		<< fSize
		<< " bytes"
		<< punt;

	/*	Construct the event editor. */

	fstream in;
	SaveTree::OpenMode = forBinaryIO;
	QuestTrigger game(new SaveTree(&in, infile), plots);

	/*	Make it so and just leave. */

	int reply = game.changeEvent();
	in.flush();
	in.close();
	Die(!reply);
}

